@@ -26,11 +26,16 @@
 #include "kernel/gpu/mig_mgr/kernel_mig_manager.h"
 #include "utils/nvassert.h"
 #include "core/locks.h"
+#include "gpu/mem_mgr/mem_mgr.h"
 
 #include "kernel/gpu/mem_mgr/ce_utils_sizes.h"
 #include "kernel/gpu/mem_mgr/channel_utils.h"
+#include "class/clcba2.h"
 
 
+#define SEC2_WL_METHOD_ARRAY_SIZE 16
+#define SHA_256_HASH_SIZE_BYTE  32
+
 /* Static helper functions */
 static void channelPushAddress(OBJCHANNEL *pChannel, CHANNEL_PB_INFO *pChannelPbInfo, NvU32 **ppPtr);
 static NvU32 channelPushMemoryProperties(OBJCHANNEL *pChannel, CHANNEL_PB_INFO *pChannelPbInfo, NvU32 **ppPtr);
@@ -61,7 +66,7 @@ channelSetupIDs
                           serverutilGenResourceHandle(pChannel->hClient, &pChannel->errNotifierIdPhys));
 
     NV_CHECK_OK_OR_RETURN(LEVEL_ERROR,
-                          serverutilGenResourceHandle(pChannel->hClient, &pChannel->copyObjectId));
+                          serverutilGenResourceHandle(pChannel->hClient, &pChannel->engineObjectId));
 
     NV_CHECK_OK_OR_RETURN(LEVEL_ERROR,
                           serverutilGenResourceHandle(pChannel->hClient, &pChannel->eventId));
@@ -96,20 +101,32 @@ channelSetupChannelBufferSizes
     NV_ASSERT_OR_RETURN_VOID(pChannel != NULL);
     NV_ASSERT_OR_RETURN_VOID(pChannel->type < MAX_CHANNEL_TYPE);
 
-    pChannel->channelNotifierSize       = CE_CHANNEL_NOTIFIER_SIZE;
-    pChannel->channelNumGpFifioEntries  = CE_NUM_COPY_BLOCKS;
-    pChannel->methodSizePerBlock        = CE_METHOD_SIZE_PER_BLOCK;
-
-    if (pChannel->type == FAST_SCRUBBER_CHANNEL)
+    if (pChannel->type == SWL_SCRUBBER_CHANNEL)
     {
-        pChannel->methodSizePerBlock = FAST_SCRUBBER_METHOD_SIZE_PER_BLOCK;
+        pChannel->methodSizePerBlock = SEC2_METHOD_SIZE_PER_BLOCK;
     }
+    else
+    {
+        pChannel->methodSizePerBlock = CE_METHOD_SIZE_PER_BLOCK;
+        if (pChannel->type == FAST_SCRUBBER_CHANNEL)
+        {
+            pChannel->methodSizePerBlock = FAST_SCRUBBER_METHOD_SIZE_PER_BLOCK;
+        }
+    }
+    pChannel->channelNotifierSize = CHANNEL_NOTIFIER_SIZE;
+    pChannel->channelNumGpFifioEntries = NUM_COPY_BLOCKS;
 
     // These sizes depend on size of each block of pushbuffer methods
-    pChannel->channelPbSize = pChannel->methodSizePerBlock * CE_NUM_COPY_BLOCKS;
-    pChannel->channelSize = pChannel->channelPbSize + CE_GPFIFO_SIZE + CE_CHANNEL_SEMAPHORE_SIZE;
-    pChannel->semaOffset = pChannel->channelPbSize + CE_GPFIFO_SIZE;
-    pChannel->finishPayloadOffset = pChannel->semaOffset + 4;
+    pChannel->channelPbSize = pChannel->methodSizePerBlock * NUM_COPY_BLOCKS;
+    pChannel->channelSize = pChannel->channelPbSize + GPFIFO_SIZE + 
+                            CHANNEL_HOST_SEMAPHORE_SIZE + CHANNEL_ENGINE_SEMAPHORE_SIZE;
+    if (pChannel->type == SWL_SCRUBBER_CHANNEL)
+    {
+        pChannel->channelSize = pChannel->channelSize + SEC2_AUTH_TAG_BUF_SEMAPHORE_SIZE;
+    }
+    pChannel->semaOffset = pChannel->channelPbSize + GPFIFO_SIZE;
+    pChannel->finishPayloadOffset = pChannel->semaOffset + CHANNEL_HOST_SEMAPHORE_SIZE;
+    pChannel->authTagBufSemaOffset = pChannel->finishPayloadOffset + CHANNEL_ENGINE_SEMAPHORE_SIZE;
 }
 
 NvU32
@@ -154,6 +171,53 @@ channelReadChannelMemdesc
     return result;
 }
 
+/*!
+ * Service any interrupts which may block the CE or SEC2 scrubber associated
+ * with this channel from making forward progress.  That includes interrupts
+ * from the CE/SEC2 engine as well as host/esched that is processing the
+ * pushbuffer.
+ */
+void
+channelServiceScrubberInterrupts(
+    OBJCHANNEL *pChannel
+)
+{
+    OBJGPU *pGpu = pChannel->pGpu;
+    Intr *pIntr = GPU_GET_INTR(pGpu);
+    KernelFifo *pKernelFifo = GPU_GET_KERNEL_FIFO(pGpu);
+    const NvU32 ceId = pChannel->ceId;
+    MC_ENGINE_BITVECTOR engines;
+    NvU32 mcIndex;
+    NvU32 engineType;
+
+    if (pChannel->type == SWL_SCRUBBER_CHANNEL)
+    {
+        mcIndex = MC_ENGINE_IDX_SEC2;
+        engineType = RM_ENGINE_TYPE_SEC2;
+    }
+    else
+    {
+        mcIndex = MC_ENGINE_IDX_CE(pChannel->ceId);
+        engineType = RM_ENGINE_TYPE_COPY(ceId);
+    }
+
+    bitVectorClrAll(&engines);
+    bitVectorSet(&engines, mcIndex);
+    if (kfifoIsHostEngineExpansionSupported(pKernelFifo))
+    {
+        NvU32 runlistId;
+        NV_ASSERT_OK(kfifoEngineInfoXlate_HAL(pGpu, pKernelFifo,
+            ENGINE_INFO_TYPE_RM_ENGINE_TYPE, engineType,
+            ENGINE_INFO_TYPE_RUNLIST, &runlistId));
+        bitVectorSet(&engines, MC_ENGINE_IDX_ESCHEDn(runlistId));
+    }
+    else
+    {
+        bitVectorSet(&engines, MC_ENGINE_IDX_FIFO);
+    }
+    intrServiceStallList_HAL(pGpu, pIntr, &engines, NV_FALSE);
+}
+
 NV_STATUS
 channelWaitForFinishPayload
 (
@@ -183,8 +247,7 @@ channelWaitForFinishPayload
 
         if (rmDeviceGpuLockIsOwner(pGpu->gpuInstance))
         {
-            Intr *pIntr = GPU_GET_INTR(pGpu);
-            intrServiceStallSingle_HAL(pGpu, pIntr, MC_ENGINE_IDX_CE(pChannel->ceId), NV_FALSE);
+            channelServiceScrubberInterrupts(pChannel);
         }
         else
         {
@@ -238,8 +301,7 @@ channelWaitForFreeEntry
         
         if (rmDeviceGpuLockIsOwner(pGpu->gpuInstance))
         {
-            Intr *pIntr = GPU_GET_INTR(pGpu);
-            intrServiceStallSingle_HAL(pGpu, pIntr, MC_ENGINE_IDX_CE(pChannel->ceId), NV_FALSE);
+            channelServiceScrubberInterrupts(pChannel);
         }
         else
         {
@@ -503,13 +565,48 @@ channelFillPbFastScrub
     return methodSize;
 }
 
+static void
+channelAddHostSema
+(
+    OBJCHANNEL *pChannel,
+    NvU32       putIndex, 
+    NvU32     **ppPtr
+)
+{
+    NvU32  *pPtr = *ppPtr;
+    NvU32  data       = 0;
+    NvU64  pSemaAddr  = 0;
+
+    // Release HOST semaphore after after gpfifo/pb segments are processed by esched
+    data = DRF_DEF(906F, _SEMAPHORED, _OPERATION, _RELEASE) |
+           DRF_DEF(906F, _SEMAPHORED, _RELEASE_SIZE, _4BYTE) |
+           DRF_DEF(906F, _SEMAPHORED, _RELEASE_WFI, _DIS);
+
+    pSemaAddr = (pChannel->pbGpuVA + pChannel->semaOffset);
+
+    //
+    // This should always be at the bottom the push buffer segment, since this
+    // denotes that HOST has read all the methods needed for this memory operation
+    // and safely assume that this GPFIFO and PB entry can be reused.
+    //
+    NV_PUSH_INC_4U(RM_SUBCHANNEL, NV906F_SEMAPHOREA,
+                   DRF_NUM(906F, _SEMAPHOREA_OFFSET, _UPPER, NvU64_HI32(pSemaAddr)),
+                   NV906F_SEMAPHOREB,
+                   DRF_NUM(906F, _SEMAPHOREB_OFFSET, _LOWER, NvU64_LO32(pSemaAddr) >> 2),
+                   NV906F_SEMAPHOREC,
+                   putIndex,
+                   NV906F_SEMAPHORED, 
+                   data);
+    *ppPtr = pPtr;
+}
+
 /** single helper function to fill the push buffer with the methods needed for
  *  memsetting using CE. This function is much more efficient in the sense it
  *  decouples the mem(set/copy) operation from managing channel resources.
  **/
 
 NvU32
-channelFillPb
+channelFillCePb
 (
     OBJCHANNEL      *pChannel,
     NvU32            putIndex, 
@@ -522,8 +619,6 @@ channelFillPb
     NvU32 *pPtr       = (NvU32 *)((NvU8 *)pChannel->pbCpuVA + (putIndex * pChannel->methodSizePerBlock));
     NvU32 *pStartPtr  = pPtr;
     NvU32  semaValue  = 0;
-    NvU32  data       = 0;
-    NvU64  pSemaAddr  = 0;
 
     NV_PRINTF(LEVEL_INFO, "PutIndex: %x, PbOffset: %x\n", putIndex, putIndex * pChannel->methodSizePerBlock);
 
@@ -554,32 +649,174 @@ channelFillPb
     // Side effect - pushed LAUNCH_DMA methods
     channelPushMethod(pChannel, pChannelPbInfo, bPipelined, bInsertFinishPayload, launchType, semaValue, &pPtr);
 
-    // Release HOST semaphore after work is completed by CE
-    data = DRF_DEF(906F, _SEMAPHORED, _OPERATION, _RELEASE) |
-           DRF_DEF(906F, _SEMAPHORED, _RELEASE_SIZE, _4BYTE) |
-           DRF_DEF(906F, _SEMAPHORED, _RELEASE_WFI, _DIS);
-
-    pSemaAddr = (pChannel->pbGpuVA + pChannel->semaOffset);
-
-    //
-    // This should always be at the bottom the push buffer segment, since this
-    // denotes that HOST has read all the methods needed for this memory operation
-    // and safely assume that this GPFIFO and PB entry can be reused.
-    //
-    NV_PUSH_INC_4U(RM_SUBCHANNEL, NV906F_SEMAPHOREA,
-                   DRF_NUM(906F, _SEMAPHOREA_OFFSET, _UPPER, NvU64_HI32(pSemaAddr)),
-                   NV906F_SEMAPHOREB,
-                   DRF_NUM(906F, _SEMAPHOREB_OFFSET, _LOWER, NvU64_LO32(pSemaAddr) >> 2),
-                   NV906F_SEMAPHOREC,
-                   putIndex,
-                   NV906F_SEMAPHORED, 
-                   data);
+    channelAddHostSema(pChannel, putIndex, &pPtr);
 
     NvU32 methodSize = (NvU32)((NvU8*)pPtr - (NvU8*)pStartPtr);
     NV_ASSERT_OR_RETURN(methodSize <= pChannel->methodSizePerBlock, 0);
     return methodSize;
 }
 
+static NV_STATUS
+addMethodsToMethodBuf
+(
+    NvU32 command,
+    NvU32 data,
+    NvU32 *pMethodBuf,
+    NvU32  index
+)
+{
+
+    NV_ASSERT_OR_RETURN((index < SEC2_WL_METHOD_ARRAY_SIZE), NV_ERR_INVALID_ARGUMENT);
+    pMethodBuf[index*2 + 0] = command;
+    pMethodBuf[index*2 + 1] = data;
+    return NV_OK;
+}
+
+/** single helper function to fill the push buffer with the methods needed for
+ *  memsetting using SEC2. This function is much more efficient in the sense it
+ *  decouples the mem(set/copy) operation from managing channel resources.
+ **/
+NV_STATUS
+channelFillSec2Pb
+(
+    OBJCHANNEL        *pChannel,
+    NvU32              putIndex,
+    NvBool             bInsertFinishPayload,
+    CHANNEL_PB_INFO   *pChannelPbInfo,
+    CCSL_CONTEXT      *pCcslCtx,
+    MEMORY_DESCRIPTOR *pAuthTagBufMemDesc,
+    MEMORY_DESCRIPTOR *pSemaMemDesc,
+    NvU64              scrubMthdAuthTagBufGpuVA,
+    NvU32              scrubAuthTagBufIndex,
+    NvU64              semaMthdAuthTagBufGpuVA,
+    NvU32              semaAuthTagBufIndex,
+    NvU32             *pMethodLength
+)
+{
+    NvU32          *pPtr                  = NULL;
+    NvU32          *pStartPtr             = NULL;
+    NvU32           execute               = 0;
+    NvU32           methodIdx             = 0;
+    NvU32          *pMethods              = NULL;
+    NV_STATUS       status                = NV_OK;
+    MemoryManager  *pMemoryManager        = NULL;
+    NvU32           scrubAuthTagBufoffset = scrubAuthTagBufIndex * SHA_256_HASH_SIZE_BYTE;
+    NvU32           semaAuthTagBufoffset  = semaAuthTagBufIndex * SHA_256_HASH_SIZE_BYTE;
+
+    NV_ASSERT_OR_RETURN((pChannel != NULL), NV_ERR_INVALID_ARGUMENT);
+    NV_ASSERT_OR_RETURN((pChannelPbInfo != NULL), NV_ERR_INVALID_ARGUMENT);
+    NV_ASSERT_OR_RETURN((pCcslCtx != NULL), NV_ERR_INVALID_ARGUMENT);
+    NV_ASSERT_OR_RETURN((pAuthTagBufMemDesc != NULL), NV_ERR_INVALID_ARGUMENT);
+    NV_ASSERT_OR_RETURN((pSemaMemDesc != NULL), NV_ERR_INVALID_ARGUMENT);
+    NV_ASSERT_OR_RETURN((pMethodLength != NULL), NV_ERR_INVALID_ARGUMENT);
+
+    pPtr = (NvU32 *)((NvU8 *)pChannel->pbCpuVA + (putIndex * pChannel->methodSizePerBlock));
+    pStartPtr = pPtr;
+    pMemoryManager = GPU_GET_MEMORY_MANAGER(pChannel->pGpu);
+
+    NV_PRINTF(LEVEL_INFO, "PutIndex: %x, PbOffset: %x\n", putIndex, putIndex * pChannel->methodSizePerBlock);
+    NvU32 transferFlags = (TRANSFER_FLAGS_USE_BAR1     |
+                           TRANSFER_FLAGS_SHADOW_ALLOC |
+                           TRANSFER_FLAGS_SHADOW_INIT_MEM);
+    NvU8* pScrubMethdAuthTagBuf = (NvU8 *)memmgrMemDescBeginTransfer(pMemoryManager, pAuthTagBufMemDesc, transferFlags);
+    NV_ASSERT_OR_RETURN((pScrubMethdAuthTagBuf != NULL), NV_ERR_INVALID_STATE);
+
+    NvU8* pSemaAuthTagBuf = (NvU8 *)memmgrMemDescBeginTransfer(pMemoryManager, pSemaMemDesc, transferFlags);
+    NV_ASSERT_OR_RETURN((pSemaAuthTagBuf != NULL), NV_ERR_INVALID_STATE);
+
+    pMethods = (NvU32*)portMemAllocNonPaged(sizeof(NvU32) * SEC2_WL_METHOD_ARRAY_SIZE * 2);
+    NV_ASSERT_OR_RETURN(pMethods != NULL, NV_ERR_NO_MEMORY);
+    portMemSet(pMethods, 0, (sizeof(NvU32) * SEC2_WL_METHOD_ARRAY_SIZE * 2));
+
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NV906F_SET_OBJECT, pChannel->classEngineID, pMethods, methodIdx++), cleanup);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_DECRYPT_COPY_DST_ADDR_HI, NvU64_HI32(pChannelPbInfo->dstAddr), pMethods, methodIdx++), cleanup);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_DECRYPT_COPY_DST_ADDR_LO, NvU64_LO32(pChannelPbInfo->dstAddr), pMethods, methodIdx++), cleanup);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_DECRYPT_COPY_SIZE, pChannelPbInfo->size, pMethods, methodIdx++), cleanup);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_METHOD_STREAM_AUTH_TAG_ADDR_HI, NvU64_HI32(scrubMthdAuthTagBufGpuVA + scrubAuthTagBufoffset), pMethods, methodIdx++), cleanup);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_METHOD_STREAM_AUTH_TAG_ADDR_LO, NvU64_LO32(scrubMthdAuthTagBufGpuVA + scrubAuthTagBufoffset), pMethods, methodIdx++), cleanup);
+
+    // Use this semaphore release to track the current method stream auth tag buffer being used
+    execute |= FLD_SET_DRF(CBA2, _EXECUTE, _NOTIFY, _ENABLE, execute);
+    execute |= FLD_SET_DRF(CBA2, _EXECUTE, _NOTIFY_ON, _END, execute);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_SEMAPHORE_A, NvU64_HI32(pChannel->pbGpuVA + pChannel->authTagBufSemaOffset), pMethods, methodIdx++), cleanup);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_SEMAPHORE_B, NvU64_LO32(pChannel->pbGpuVA + pChannel->authTagBufSemaOffset), pMethods, methodIdx++), cleanup);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_SET_SEMAPHORE_PAYLOAD_LOWER, scrubAuthTagBufIndex, pMethods, methodIdx++), cleanup);
+
+    execute |= FLD_SET_DRF(CBA2, _EXECUTE, _PHYSICAL_SCRUBBER, _ENABLE, execute);
+    NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_EXECUTE, execute, pMethods, methodIdx++), cleanup);
+
+    NvU32 hmacBufferSizeBytes = 2 * methodIdx * sizeof(NvU32); 
+    NvU8  hmacDigest[SHA_256_HASH_SIZE_BYTE] = {0};
+    NvU8* pBufScrub = &pScrubMethdAuthTagBuf[scrubAuthTagBufoffset];
+
+    ccslSign(pCcslCtx, hmacBufferSizeBytes, (NvU8 *)pMethods, hmacDigest);
+
+    portMemCopy((void*)pBufScrub, SHA_256_HASH_SIZE_BYTE, (const void*)&hmacDigest[0], SHA_256_HASH_SIZE_BYTE);
+
+    if (methodIdx > SEC2_WL_METHOD_ARRAY_SIZE)
+    {
+        status = NV_ERR_INVALID_ARGUMENT;
+        goto cleanup;
+    }
+
+    for (NvU32 i = 0; i < methodIdx; i++)
+    {
+        NV_PUSH_INC_1U(RM_SUBCHANNEL, pMethods[i*2 + 0], pMethods[i*2 + 1]);
+    }
+
+    // insert standalone semaphore to track sub operation completion
+    if (bInsertFinishPayload)
+    {
+        NvU32 semaD = 0;
+        methodIdx = 0;
+        portMemSet(pMethods, 0, (sizeof(NvU32) * SEC2_WL_METHOD_ARRAY_SIZE * 2));
+        semaD |= FLD_SET_DRF(CBA2, _SEMAPHORE_D, _FLUSH_DISABLE, _FALSE, execute);
+        NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_METHOD_STREAM_AUTH_TAG_ADDR_HI, NvU64_HI32(semaMthdAuthTagBufGpuVA + semaAuthTagBufoffset), pMethods, methodIdx++), cleanup);
+        NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_METHOD_STREAM_AUTH_TAG_ADDR_LO, NvU64_LO32(semaMthdAuthTagBufGpuVA + semaAuthTagBufoffset), pMethods, methodIdx++), cleanup);
+        NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_SEMAPHORE_A, NvU64_HI32(pChannel->pbGpuVA + pChannel->finishPayloadOffset), pMethods, methodIdx++), cleanup);
+        NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_SEMAPHORE_B, NvU64_LO32(pChannel->pbGpuVA + pChannel->finishPayloadOffset), pMethods, methodIdx++), cleanup);
+        NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_SET_SEMAPHORE_PAYLOAD_LOWER, pChannelPbInfo->payload, pMethods, methodIdx++), cleanup);
+        NV_ASSERT_OK_OR_GOTO(status, addMethodsToMethodBuf(NVCBA2_SEMAPHORE_D, semaD, pMethods, methodIdx++), cleanup);
+
+        hmacBufferSizeBytes = 2 * methodIdx * sizeof(NvU32); 
+        portMemSet(&hmacDigest[0], 0, SHA_256_HASH_SIZE_BYTE);
+        NvU8* pBufSema = &pSemaAuthTagBuf[semaAuthTagBufoffset];
+
+        ccslSign(pCcslCtx, hmacBufferSizeBytes, (NvU8 *)pMethods, hmacDigest);
+
+        portMemCopy((void*)pBufSema, SHA_256_HASH_SIZE_BYTE, (const void*)&hmacDigest[0], SHA_256_HASH_SIZE_BYTE);
+
+        if (methodIdx > SEC2_WL_METHOD_ARRAY_SIZE)
+        {
+            status = NV_ERR_INVALID_ARGUMENT;
+            goto cleanup;
+        }
+
+        for (NvU32 i = 0; i < methodIdx; i++)
+        {
+            NV_PUSH_INC_1U(RM_SUBCHANNEL, pMethods[i*2 + 0], pMethods[i*2 + 1]);
+        }        
+    }
+
+    channelAddHostSema(pChannel, putIndex, &pPtr);
+
+    *pMethodLength = 0;
+    NvU32 methodSize = (NvU32)((NvU8*)pPtr - (NvU8*)pStartPtr);
+    NV_ASSERT_OR_RETURN(methodSize <= pChannel->methodSizePerBlock, NV_ERR_INVALID_STATE);
+    *pMethodLength = methodSize;
+
+cleanup:
+    if (pSemaAuthTagBuf != NULL)
+    {
+        memmgrMemDescEndTransfer(pMemoryManager, pSemaMemDesc, transferFlags);
+    }
+    if (pScrubMethdAuthTagBuf != NULL)
+    {
+        memmgrMemDescEndTransfer(pMemoryManager, pAuthTagBufMemDesc, transferFlags);
+    }
+    portMemFree(pMethods);
+    return status;
+}
 
 /*** Implementation for static methods ***/
 static NvU32 
